﻿/******************************************************************************
 * Copyright 2019 The siasun Transport Authors. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 1.0 (the "License");
 *****************************************************************************/

#include "canbus.h"
#include"candatatype.h"
#include"messagecoder.h"
#include"../common/messagetopics.h"
#include "std_msgs/String.h"

namespace  robot{
namespace canbus {



std::string Canbus::Name() const { return "canbus"; }

int Canbus::Init() {
  LOG(INFO)<< Name()<<" Init start...";
  node_handle_.reset(new ros::NodeHandle());

  //load libpcan.so using dlopen function,return handle for further use
  libm_handle = dlopen("libpcan.so", RTLD_LAZY );
  if (!libm_handle){
      LOG(ERROR)<< Name()<<" Init Open Error:" << dlerror();//if file can't be loaded,return null,get reason using dlerror function
      return ROBOT_FAILTURE;
  }
  char *errorInfo;//error information pointer
  //one-to-one mapping using dlsym function,if return null,mapping would be failed
  funCAN_Init         =(funCAN_Init_TYPE)         dlsym(libm_handle,"CAN_Init");
  funLINUX_CAN_Open   =(funLINUX_CAN_Open_TYPE)   dlsym(libm_handle,"LINUX_CAN_Open");
  funCAN_Close        =(funCAN_Close_TYPE)        dlsym(libm_handle,"CAN_Close");
  funCAN_VersionInfo  =(funCAN_VersionInfo_TYPE)  dlsym(libm_handle,"CAN_VersionInfo");
  funLINUX_CAN_Read   =(funLINUX_CAN_Read_TYPE)   dlsym(libm_handle,"LINUX_CAN_Read");
  funCAN_Status       =(funCAN_Status_TYPE)       dlsym(libm_handle,"CAN_Status");
  funnGetLastError    =(funnGetLastError_TYPE)    dlsym(libm_handle,"nGetLastError");
  funCAN_Write        =(funCAN_Write_TYPE)       dlsym(libm_handle, "CAN_Write");

  errorInfo = dlerror();//get error using dlerror function,and clear the error list in memory
  if (errorInfo != NULL){
      LOG(ERROR)<<Name()<<" Init Dlsym Error:"<<errorInfo;
      return ROBOT_FAILTURE;
  }

  //initialize the VehHBrkStat as Hold
  status1.uiVehHBrkStat = 1;
  LOG(INFO)<<Name()<<" init end.";
  return ROBOT_SUCCESS;
}

//打开Can总线
int Canbus::Start() {
  char chVersion[VERSIONSTRING_LEN];            //store information of can version
  const char  *szDevNode = DEFAULT_NODE;  //define const pointer point to device name
  pcan_handle = funLINUX_CAN_Open(szDevNode, O_RDWR );//use mapping function

  //judge whether the call is success.if pcan_handle=null,the call would be failed
  if(pcan_handle){
      LOG(INFO)<<Name()<<": "<<szDevNode<<" opened";
      errno = funCAN_VersionInfo(pcan_handle, chVersion);
      if (!errno)
          LOG(INFO)<<Name()<<": driver version ="<<chVersion;
      else {
          LOG(ERROR)<<Name()<<": CAN_VersionInfo() error";
      }
      if (unBaudrate) {
              errno = funCAN_Init(pcan_handle, unBaudrate, nExtended);
              if (errno) {
                  LOG(ERROR)<<Name()<<": CAN_Init()";
              }
              else
                  LOG(INFO)<<" Device Info: " << szDevNode <<"; CAN_BAUD_1M; CAN_INIT_TYPE_ST";
          }
  }
  else
  {
      LOG(INFO)<<Name()<<": can't open ."<<szDevNode;
  }

  ros::Rate loop_rate(2000);
  // 5. set timer to triger publish info periodly

  timer_ = CreateTimer(ros::Duration(duration), &Canbus::OnTimer, this);
   //data receive
   while(ros::ok())
   {
       read_loop();
       ros::spinOnce();
       loop_rate.sleep();
   }
}


void Canbus::OnTimer(const ros::TimerEvent &) {
 // LOG(INFO)<< Name()<< " timer start ..";
  if(heatbeat > 15)
  {
      heatbeat = 0;
  }
  heatbeat ++;
  icuMonitor.uiICUAlive = heatbeat;
  MessageCoder::CanMsgPackage((void*)&icuMonitor,0x5FA,&canIcuMonitor);
  MessageCoder::CanMsgPackage((void*)&status1,0x6F1,&canStatus1);
  MessageCoder::CanMsgPackage((void*)&status2,0x6F2,&canStatus2);
  MessageCoder::CanMsgPackage((void*)&status3,0x6F3,&canStatus3);
  write_data(&canStatus1);
  write_data(&canStatus2);
  write_data(&canStatus3);
  write_data(&canIcuMonitor);
}

void Canbus::Stop() {

  //Must close h handle firstly,then close file using dlclose
  if (pcan_handle) {
      funCAN_Close(pcan_handle);
  }
  LOG(INFO)<<Name()<<": closed. ";
  ros::shutdown();
    LOG(INFO)<<Name()<<": node shutdown. ";
}


//read from CAN forever - until manual break
int Canbus::read_loop()
{
    TPCANRdMsg msg;
    __u32 status;

    if (funLINUX_CAN_Read(pcan_handle, &msg)) {
        LOG(ERROR)<<Name()<<" receive: LINUX_CAN_Read() error.";
        return errno;
    }
    read_data(&(msg.Msg));
     // deal receive data
    // check if a CAN status is pending
    if (msg.Msg.MSGTYPE & MSGTYPE_STATUS) {
        status = funCAN_Status(pcan_handle);
        if ((int)status < 0) {
            errno = funnGetLastError();
            LOG(ERROR)<<Name()<<" receive: CAN_Status() error.";
            return errno;
        }
    }

    return ROBOT_SUCCESS;
}

int Canbus::read_data(TPCANMsg *pMsgBuff)
{
//    char strId[4] = {0};
//    sprintf(strId,"%3xh",pMsgBuff->ID);
//    char strData[16]={0};
//    MessageCoder::ConvertUnCharToStr(strData,pMsgBuff->DATA,8);

//    LOG(INFO)<<Name()<<" Receive CanData : ID: "<<std::string(strId)<<",Data: "<<std::string(strData);
    if(pMsgBuff->ID == 0x1A1)
    {
//       char strId1[4] = {0};
//       char strData1[16]={0};
//       sprintf(strId1,"%3xh",pMsgBuff->ID);
//       MessageCoder::ConvertUnCharToStr(strData1,pMsgBuff->DATA,8);
//       LOG(INFO)<<Name()<<" Receive CanData : ID: "<<std::string(strId1)<<",Data: "<<std::string(strData1);
       T_OBD_ACC_PSNG tOBDAccPsng = *((T_OBD_ACC_PSNG*)MessageCoder::CanMsgAnalyse(0x1A1,pMsgBuff->DATA));
       status1.uiAccPsng = tOBDAccPsng.obdAccPsng;
//       LOG(INFO)<<Name()<<"ID:0x1A1,Data: "<<tOBDAccPsng.obdAccPsng;
    }
    else if(pMsgBuff->ID == 0x0F1)
    {
//       char strId2[4] = {0};
//       char strData2[16]={0};
//       sprintf(strId2,"%3xh",pMsgBuff->ID);
//       MessageCoder::ConvertUnCharToStr(strData2,pMsgBuff->DATA,8);
//       LOG(INFO)<<Name()<<" Receive CanData : ID: "<<std::string(strId2)<<",Data: "<<std::string(strData2);
       T_OBD_BRK_PSNG tOBDBrkPsng = *((T_OBD_BRK_PSNG*)MessageCoder::CanMsgAnalyse(0x0F1,pMsgBuff->DATA));
       status1.uiBrkPsng = tOBDBrkPsng.obdBrkPsng;
//       LOG(INFO)<<Name()<<"ID:0x0F1,Data: "<<tOBDBrkPsng.obdBrkPsng;
    }
    else if(pMsgBuff->ID == 0x135)
    {
//       char strId3[4] = {0};
//       char strData3[16]={0};
//       sprintf(strId3,"%3xh",pMsgBuff->ID);
//       MessageCoder::ConvertUnCharToStr(strData3,pMsgBuff->DATA,8);
//       LOG(INFO)<<Name()<<" Receive CanData : ID: "<<std::string(strId3)<<",Data: "<<std::string(strData3);
       T_OBD_SFT_PSNG tOBDSftPsng = *((T_OBD_SFT_PSNG*)MessageCoder::CanMsgAnalyse(0x135,pMsgBuff->DATA));
       status1.uiSftPsng = tOBDSftPsng.obdSftPsng;
//      LOG(INFO)<<Name()<<"ID:0x135,Data: "<<tOBDSftPsng.obdSftPsng;
    }
    else if(pMsgBuff->ID == 0x1F1)
    {
//        char strId4[4] = {0};
//        char strData4[16]={0};
//        sprintf(strId4,"%3xh",pMsgBuff->ID);
//        MessageCoder::ConvertUnCharToStr(strData4,pMsgBuff->DATA,8);
//        LOG(INFO)<<Name()<<" Receive CanData : ID: "<<std::string(strId4)<<",Data: "<<std::string(strData4);
        T_OBD_VEH_BRK_STAT tOBDVehBrkStat = *((T_OBD_VEH_BRK_STAT*)MessageCoder::CanMsgAnalyse(0x1F1,pMsgBuff->DATA));
        status1.uiVehHBrkStat = tOBDVehBrkStat.obdVehHBrkStat;
//        LOG(INFO)<<Name()<<"ID:0x1F1,Data: "<<tOBDVehBrkStat.obdVehHBrkStat;
    }
    else if(pMsgBuff->ID == 0x121)
    {
//        char strId5[4] = {0};
//        char strData5[16]={0};
//        sprintf(strId5,"%3xh",pMsgBuff->ID);
//        MessageCoder::ConvertUnCharToStr(strData5,pMsgBuff->DATA,8);
//        LOG(INFO)<<Name()<<" Receive CanData : ID: "<<std::string(strId5)<<",Data: "<<std::string(strData5);
        T_OBD_VEH_SW_STAT tOBDVehSwStat = *((T_OBD_VEH_SW_STAT*)MessageCoder::CanMsgAnalyse(0x121,pMsgBuff->DATA));
        status1.uiVehSwStat = tOBDVehSwStat.obdVehSwStat;
//        LOG(INFO)<<Name()<<"ID:0x121,Data: "<<tOBDVehSwStat.obdVehSwStat;
    }
    else if(pMsgBuff->ID == 0x1E5)
    {
//       char strId6[4] = {0};
//       char strData6[16]={0};
//       sprintf(strId6,"%3xh",pMsgBuff->ID);
//       MessageCoder::ConvertUnCharToStr(strData6,pMsgBuff->DATA,8);
//       LOG(INFO)<<Name()<<" Receive CanData : ID: "<<std::string(strId6)<<",Data: "<<std::string(strData6);
       T_OBD_STR_ANG tOBDStrAng = *((T_OBD_STR_ANG*)MessageCoder::CanMsgAnalyse(0x1E5,pMsgBuff->DATA));
       status2.iStrAng = tOBDStrAng.obdStrAng;
//        LOG(INFO)<<Name()<<"ID:0x1E5,Data: "<<tOBDStrAng.obdStrAng;
    }
    else if(pMsgBuff->ID == 0x0C9)
    {
//        char strId7[4] = {0};
//        char strData7[16]={0};
//        sprintf(strId7,"%3xh",pMsgBuff->ID);
//        MessageCoder::ConvertUnCharToStr(strData7,pMsgBuff->DATA,8);
//        LOG(INFO)<<Name()<<" Receive CanData : ID: "<<std::string(strId7)<<",Data: "<<std::string(strData7);
        T_OBD_ENG_SPD tOBDEngSpd = *((T_OBD_ENG_SPD*)MessageCoder::CanMsgAnalyse(0x0C9,pMsgBuff->DATA));
        status2.uiEngSpd = tOBDEngSpd.obdEngSpd;
        //LOG(INFO)<<Name()<<"ID:0x0C9,Data: "<<tOBDEngSpd.obdEngSpd;
    }
    else if((pMsgBuff->ID > 0x5F0 && pMsgBuff->ID < 0x7FF) || pMsgBuff->ID ==0x2E0 || pMsgBuff->ID ==0x2E1)
    {
        //publish_data(pMsgBuff);

    }
    return ROBOT_SUCCESS;
}

void Canbus::publish_data(TPCANMsg *pMsgBuff)
{
    //dataAnalysis::CanData canData;
    //canData.set_id(pMsgBuff->ID);
    //canData.set_data(reinterpret_cast<const char*>(pMsgBuff->DATA));
    std_msgs::String ros_msg;
    std::stringstream ss;
    ss << "Id:" << pMsgBuff->ID << ",Data:" << pMsgBuff->DATA;
    ros_msg.data = ss.str();
    ros::Publisher canData_pub = node_handle_->advertise<std_msgs::String>(robot::common::CanDataAnalysis, 200);
    canData_pub.publish(ros_msg);
    LOG(INFO)<<Name()<<" Publish CanData :"<<ss.str();
}

int Canbus::write_data(TPCANMsg *pMsgBuff)
{
    char strId[4] = {0};
    sprintf(strId,"%3xh",pMsgBuff->ID);
    char strData[16]={0};
    MessageCoder::ConvertUnCharToStr(strData,pMsgBuff->DATA,8);
  LOG(INFO)<<Name()<<" send CanData : ID: "<<std::string(strId)<<",Data:"<<std::string(strData);
  if (funCAN_Write(pcan_handle, pMsgBuff)){
     LOG(ERROR)<<Name()<< " write: LINUX_CAN_write_timeout error. ";
     return errno;
  }
   return ROBOT_SUCCESS;
}


}  // namespace canbus
}  // namespace autopilot
